"""Implements a series of named buttons that can be accessed"""


import serge.actor
import serge.events


class ButtonUI(serge.actor.CompositeActor):
    """Implements a series of named buttons"""

    def __init__(self, tag, name, button_setup, callback=None):
        """Initialise the ui"""
        super(ButtonUI, self).__init__(tag, name)
        #
        self.button_setup = button_setup
        self.buttons = {}
        self.callback = callback

    def addedToWorld(self, world):
        """Added the actor to the world"""
        super(ButtonUI, self).addedToWorld(world)
        #
        for name, (x, y), (on_sprite, off_sprite), state in self.button_setup:
            btn = world.addActor(Button('btn', 'btn-%s' % name, on_sprite, off_sprite))
            btn.setState(state)
            btn.moveTo(x, y)
            if self.callback:
                btn.linkEvent(serge.events.E_LEFT_CLICK, lambda o, arg: self.callback(arg), name)
            self.buttons[name] = btn

    def getButton(self, name):
        """Return the named button"""
        return self.buttons[name]

    def setStates(self, state, names=None):
        """Set the states of a number of buttons"""
        if names is None:
            names = self.buttons.keys()
        for name in names:
            self.buttons[name].setState(state)

    def toggleStates(self, names=None):
        """Toggle the states of a number of buttons"""
        if names is None:
            names = self.buttons.keys()
        for name in names:
            self.buttons[name].toggleState()

    def setLayerName(self, name):
        """Set the layer for this actor"""
        for btn in self.buttons.values():
            btn.setLayerName(name)


class Button(serge.actor.Actor):
    """Implements a toggled button"""

    def __init__(self, tag, name, on_sprite, off_sprite):
        """Initialise the button"""
        super(Button, self).__init__(tag, name)
        #
        self.on_sprite = on_sprite
        self.off_sprite = off_sprite
        self._state = True
        self.setState(True)

    def setState(self, state):
        """Set the current state"""
        self.setSpriteName(self.on_sprite if state else self.off_sprite)
        self._state = state

    def getState(self):
        """Return the state"""
        return self._state

    def toggleState(self):
        """Toggle the state"""
        self.setState(not self.getState())
