"""The high scores screen for the game"""

import pygame

import serge.actor
import serge.visual
import serge.events
import serge.common
import serge.sound
import serge.blocks.utils
import serge.blocks.visualblocks
import serge.blocks.behaviours
import serge.blocks.actors
import serge.blocks.onlinescores
import serge.blocks.layout
import serge.blocks.animations

from theme import G, theme
import storage
import uielements


class HighScoreScreen(uielements.StandardUI):
    """The logic for the high score screen"""

    confirm_options = ['Continue', 'Quit']

    def __init__(self, options):
        """Initialise the screen"""
        super(HighScoreScreen, self).__init__('item', 'high-score-screen')
        self.options = options

    def addedToWorld(self, world):
        """The screen was added to the world"""
        super(HighScoreScreen, self).addedToWorld(world)
        L = theme.getTheme('high-score-screen').getProperty
        manager = serge.blocks.behaviours.BehaviourManager('behaviours', 'behaviours')
        world.addActor(manager)
        #
        self.clock_digits.setValue('....')
        self.flags_digits.setValue('....')
        #
        # Your score
        self.score = serge.blocks.utils.addTextToWorld(
            world, 'Your score', 'score', theme.getTheme('high-score-screen'), 'main'
        )
        #
        # Your best score
        self.best_score = serge.blocks.utils.addTextToWorld(
            world, 'Your best score', 'best-score', theme.getTheme('high-score-screen'), 'main'
        )
        self.best_score.visible = False
        #
        # Name
        serge.blocks.utils.addTextToWorld(
            world, 'Name', 'name', theme.getTheme('high-score-screen'), 'main'
        )
        self.name_entry = serge.blocks.utils.addActorToWorld(
            world,
            serge.blocks.actors.TextEntryWidget(
                'entry', 'entry', L('entry-width'), L('entry-height'), L('entry-colour'),
                L('entry-size'), 'terminal', show_cursor=True, justify='center',
                background_visual=serge.blocks.visualblocks.Rectangle(
                    (L('entry-width'), L('entry-height')), L('entry-fill-colour'),
                    stroke_width=L('entry-stroke-width'), stroke_colour=L('entry-stroke-colour')
                ),
                background_layer='main',
            ),
            layer_name='ui',
            center_position=L('entry-position'),
        )
        self.name_entry.setText(storage.User.getLastUser().name)
        self.name_entry.cursor_pos = len(self.name_entry.text.value)
        #
        # Clear text
        self.clear = serge.blocks.utils.addTextToWorld(
            world, 'Clear Name', 'clear', theme.getTheme('high-score-screen'), 'ui', serge.blocks.animations.AnimatedActor
        )
        self.clear.linkEvent(serge.events.E_LEFT_CLICK, self.clearName)
        self.clear.addRegisteredAnimation('pulsed-visibility')
        #
        # High scores
        self.submit = serge.blocks.utils.addTextToWorld(
            world, 'Submit Score', 'submit', theme.getTheme('high-score-screen'), 'ui', serge.blocks.animations.AnimatedActor
        )
        #self.high_scores.active = False
        self.submit.linkEvent(serge.events.E_LEFT_CLICK, self.submitScore)
        self.submit.addRegisteredAnimation('pulsed-visibility')
        self.submit.active = False
        #
        # Status
        self.status = serge.blocks.utils.addTextToWorld(
            world, 'Submitting ...', 'status', theme.getTheme('high-score-screen'), 'ui', serge.blocks.animations.AnimatedActor
        )
        self.status.active = False
        #
        # The high score table
        self.table = serge.blocks.onlinescores.AsyncOnlineScoreTable(
            L('app-url'), L('app-name'), 1
        )
        #
        self.current_game = None
        self.current_score = None
        self.current_mode = None
        self.difficulty_lights.setStates(False)
        #
        # High score tables
        difficulties = G('difficulties').keys()
        #
        # Bar for the items in the table
        bar = serge.blocks.utils.addActorToWorld(
            world,
            serge.blocks.layout.VerticalBar(
                'vbar', 'vbar', L('table-width'), item_height=L('table-item-height')
            ),
            layer_name='ui',
            center_position=L('table-layout-position')
        )
        self.table_title = bar.addActor(
            serge.blocks.actors.StringText(
                'title', 'title', difficulties[0], colour=L('table-title-colour'),
                font_size=L('table-title-size'), font_name=L('table-title-font')
            ),
            layer_name='ui',
        )
        bar.addBlanks(1)
        #
        # The actual table
        self.table_view = serge.blocks.onlinescores.SimpleHSTableView(
            'hs-table', 'hs-table', L('app-url'), L('app-name'),
            '%s - %s' % (G('menu-default'), G('mode-default')),
            storage.User.getLastUser().name, L('max-scores'),
            bar,
            theme.getTheme('high-score-screen'),
            fmt='%(num)02d ... %(player)s ... %(score)5.1f'
        )
        self.table_view.updateScores()
        self.table_view.setLayerName('ui')
        self.table_view.moveTo(*L('table-layout-position'))
        world.addActor(self.table_view)
        #
        self.navigation_lights.setStates(True)
        self.navigation_lights.getButton('new-game').linkEvent(serge.events.E_LEFT_CLICK, self.clearName)
        self.navigation_lights.getButton('high-scores').linkEvent(serge.events.E_LEFT_CLICK, serge.blocks.utils.backToPreviousWorld('screen'))
        self.navigation_lights.getButton('submit').linkEvent(serge.events.E_LEFT_CLICK, self.submitScore)
        #
        # Back link
        back = serge.blocks.utils.addVisualActorToWorld(
            world, 'button', 'back',
            serge.visual.Text('Back', L('back-colour'), font_size=L('back-font-size'), font_name='terminal'),
            'foreground', 
            center_position=L('back-position'),
            actor_class=serge.blocks.animations.AnimatedActor)
        back.linkEvent(serge.events.E_LEFT_CLICK, serge.blocks.utils.backToPreviousWorld('screen'))
        back.addRegisteredAnimation('pulsed-visibility')
        #
        # Cheating
        if self.options.cheat:
            serge.blocks.utils.addActorToWorld(
                world, serge.blocks.actors.FPSDisplay(G('fps-x'), G('fps-y'), G('fps-colour'), G('fps-size')))

    def updateActor(self, interval, world):
        """Update this actor"""
        super(HighScoreScreen, self).updateActor(interval, world)

    def setCurrentScore(self, score, game, mode):
        """Sets the current score"""
        self.current_score = score
        self.current_game = game
        self.current_mode = mode
        if score is None:
            self.submit.active = self.score.active = False
            self.clock_digits.setValue('....')
        else:
            self.submit.active = self.score.active = True
            self.clock_digits.setValue('%04d' % score)
        #
        self.navigation_lights.getButton('submit').setState(self.submit.active)
        self.setDifficulty(game, mode)
        self.status.active = False

    def submitScore(self, obj, arg):
        """Submit the high score online"""
        #
        # Watch out for trying to submit when there is no submitting
        if not self.submit.active:
            return
        #
        serge.sound.Sounds.play('select')
        self.log.info('Submitting high score')
        #
        self.navigation_lights.getButton('submit').setState(False)
        self.submit.active = False
        self.status.active = True
        self.status.visual.setText('Submitting user ...')
        player_name = self.name_entry.getText()
        result = self.table.createPlayer(player_name)
        result.add_done_callback(self.playerCreated)
        #
        storage.User.updateUser(self.name_entry.getText(), self.current_game)

    def playerCreated(self, result):
        """The player was created"""
        self.status.visual.setText('Submitting score ...')
        next_result = self.table.recordScore(
            self.name_entry.getText(),
            '%s - %s' % (self.current_game,self.current_mode),
            self.current_score
        )
        next_result.add_done_callback(self.scoreSubmitted)

    def scoreSubmitted(self, result):
        """The score was submitted"""
        try:
            text = result.result()
        except serge.blocks.onlinescores.OnlineMethodCallFailed, err:
            text = err.args[0]
        else:
            text = 'Score submitted!'
        self.status.visual.setText(text)
        self.updateScores(None, None)
        self.setDifficulty(self.current_game, self.current_mode)

    def clearName(self, obj, arg):
        """Clear the name"""
        serge.sound.Sounds.play('select')
        self.name_entry.setText('')

    def updateScores(self, obj, arg):
        """Update all the score tables"""
        self.updating = 1
        self.status.active = True
        self.table_view.player_name = self.name_entry.getText()
        self.table_view.updateScores(done=self.updateDone)

    def updateDone(self):
        """An update to a table is done"""
        self.log.debug('Update done. Updating is %s' % self.updating)
        self.updating -= 1
        if self.updating <= 0:
            self.status.visual.setText('Complete')
        else:
            self.status.visual.setText('Updating %d tables ...' % self.updating)

    def playerBestUpdate(self, result):
        """Update the player's best"""
        try:
            answer = result.result()
        except Exception, err:
            self.log.error('Failed to get player\'s best: %s' % err)
        else:
            if len(answer) == 0:
                self.log.info('No best score for this player')
            else:
                self.log.info('Got player best results %s' % (answer,))
                self.best_score.visible = True
                self.best_score.visual.setText('Your best score: %5.1f' % answer[0][1])

    def menuClick(self, obj, difficulty):
        """Clicked on the difficulty menu"""
        self.log.debug('Clicked on menu %s, %s' % (obj, difficulty))
        if hasattr(self, 'difficulty_lights'):
            serge.sound.Sounds.play('button')
            self.setDifficulty(difficulty, self.mode_menu.getSelection())

    def setDifficulty(self, difficulty, mode):
        """Set the current difficulty"""
        #
        # Change the lights
        self.difficulty_lights.setStates(False)
        self.difficulty_lights.getButton(difficulty).setState(True)
        self.difficulty_menu.selectItem(difficulty, do_callback=False)
        #
        self.mode_lights.setStates(False)
        self.mode_lights.getButton(mode).setState(True)
        self.mode_menu.selectItem(mode, do_callback=False)
        #
        # Hide the old table and make the chosen one visible
        category = '%s - %s' % (difficulty, mode)
        self.table_title.visual.setText(category)
        self.table_view.category_name = category
        self.table_view.updateScores()
        #
        # Get the player's best score also
        self.best_score.visible = False
        player_best = self.table.getScores(self.table_view.category_name, self.table_view.player_name, 1)
        player_best.add_done_callback(self.playerBestUpdate)

    def setMode(self, mode, silent=False):
        """Set the mode"""
        self.log.debug('Clicked on mode %s' % mode)
        if hasattr(self, 'difficulty_lights'):
            super(HighScoreScreen, self).setMode(mode, silent)
            #
            self.setDifficulty(self.difficulty_menu.getSelection(), mode)


def main(options):
    """Create the main logic"""
    #
    # The screen actor
    s = HighScoreScreen(options)
    world = serge.engine.CurrentEngine().getWorld('high-score-screen')
    world.addActor(s)
    #
    # The behaviour manager
    manager = serge.blocks.behaviours.BehaviourManager('behaviours', 'behaviours')
    world.addActor(manager)
    #
    # Screenshots
    if options.screenshot:
        manager.assignBehaviour(None, 
            serge.blocks.behaviours.SnapshotOnKey(key=pygame.K_s, size=G('screenshot-size')
                , overwrite=False, location='screenshots'), 'screenshots')

