"""The main screen for the game"""

import pygame
import time

import serge.actor
import serge.visual
import serge.sound
import serge.events
import serge.common
import serge.blocks.utils
import serge.blocks.visualblocks
import serge.blocks.behaviours
import serge.blocks.actors
import serge.blocks.layout
import serge.blocks.directions
import serge.blocks.animations

from theme import G, theme
import common 
import field
import uielements
import metalsensor
import oscilloscope

# IDEA: Oscilloscope that shows direction of mines
# IDEA: Make the sensor show likelihood of mine below the current square


class MainScreen(uielements.StandardUI):
    """The logic for the main screen"""

    confirm_options = ['Start New Game', 'Continue Current', 'Quit']

    def __init__(self, options):
        """Initialise the screen"""
        super(MainScreen, self).__init__('item', 'main-screen')
        self.options = options
        self.initialised = False

    def addedToWorld(self, world):
        """Added to the world"""
        super(MainScreen, self).addedToWorld(world)
        #
        # Clock
        self.clock = serge.blocks.utils.addTextToWorld(
            world, 'Time', 'clock', theme.getTheme('main'), 'main'
        )
        #
        # Flags
        self.flags = serge.blocks.utils.addTextToWorld(
            world, 'Flags', 'flags', theme.getTheme('main'), 'main'
        )
        #
        # Flag counter
        self.flag_counter = serge.blocks.utils.addActorToWorld(
            world,
            serge.actor.Actor('counter', 'flag-counter'),
            layer_name='main',
            center_position=G('flag-counter-position')
        )
        self.flag_counter.visual = serge.blocks.visualblocks.BlockedProgressBar(
            G('flag-counter-size'), G('flag-counter-block-size'), False, G('flag-counter-blocks'), True
        )
        self.flag_counter.visual.setValue(0)
        self._last_flag_placed = time.time()
        self._last_turns = 0
        # This thing isn't working very well at the moment
        self.flag_counter.active = False
        #
        # Oscilloscope
        self.scope = serge.blocks.utils.addActorToWorld(
            world,
            oscilloscope.Oscilloscope(
                'scope', 'scope',
                G('scope-size'), G('scope-colour'), G('scope-x-points'),
                G('scope-x-width'), G('scope-relaxation'), G('scope-angle-jitter'),
                G('scope-amplitude-jitter'), G('scope-y-offset')),
            layer_name='main',
            center_position=G('scope-position')
        )
        self.scope.setAngles([])
        #
        self.scope.addChild(serge.blocks.utils.addSpriteActorToWorld(
            world, 'bg', 'scope-bg', 'scope', 'background', G('scope-position')
        ))
        #
        # Success
        self.success = serge.blocks.utils.addTextToWorld(
            world, 'Well Done!', 'success', theme.getTheme('main'), 'ui'
        )
        #
        # High scores
        self.high_scores = serge.blocks.utils.addTextToWorld(
            world, 'High Scores', 'scores', theme.getTheme('main'), 'ui', serge.blocks.animations.AnimatedActor
        )
        self.high_scores.linkEvent(serge.events.E_LEFT_CLICK, self.gotoHighScores)
        self.high_scores.addRegisteredAnimation('pulsed-visibility')
        #
        # The restart button
        self.restart = serge.blocks.utils.addTextToWorld(
            world, 'New Game', 'restart', theme.getTheme('main'), 'ui', serge.blocks.animations.AnimatedActor
        )
        self.restart.linkEvent(serge.events.E_LEFT_CLICK, self.restartClick)
        self.restart.addRegisteredAnimation('pulsed-visibility')
        #
        # The credits button
        self.credits = serge.blocks.utils.addTextToWorld(
            world, 'Credits', 'credits', theme.getTheme('main'), 'ui', serge.blocks.animations.AnimatedActor
        )
        self.credits.linkEvent(serge.events.E_LEFT_CLICK, serge.blocks.utils.worldCallback('credits-screen', 'screen'))
        self.credits.addRegisteredAnimation('pulsed-visibility')
        #
        # Difficulty menu
        self.difficulties = G('difficulties')
        self._accepted_difficulty = self.difficulty_menu.getSelection()
        self.difficulty_lights.setLayerName('buttons')
        self.difficulty_lights.setStates(False)
        self.difficulty_lights.getButton(self._accepted_difficulty).setState(True)
        #
        self.navigation_lights.setLayerName('buttons')
        self.navigation_lights.setStates(False)
        self.navigation_lights.getButton('new-game').linkEvent(serge.events.E_LEFT_CLICK, self.restartClick)
        self.navigation_lights.getButton('high-scores').linkEvent(serge.events.E_LEFT_CLICK, self.gotoHighScores)
        self.navigation_lights.getButton('submit').linkEvent(
            serge.events.E_LEFT_CLICK, serge.blocks.utils.worldCallback('credits-screen', 'screen'))
        #
        self._accepted_mode = self.mode_menu.getSelection()
        #
        # Visual elements that are activated on game over
        self.game_over_elements = serge.actor.ActorCollection([
            self.high_scores, self.success, self.restart, self.credits
        ])
        self.game_over_elements.forEach().active = False
        #
        # The metal detector sensor
        self.sensor = serge.blocks.utils.addActorToWorld(
            world, metalsensor.MetalSensor('sensor', 'sensor'),
            layer_name='main',
            center_position=G('sensor-position')
        )
        self.sensor.setValue(0.0)
        self.sensor.active = common.C_GAME_MODE == common.C_MODE_NEW
        #
        # The minefield itself
        self.createMinefield()
        #
        # Variables
        self.start_time = time.time()
        self.total_flags = self.getNumberBombs()
        self.flagChanged(0, None)
        self.initialised = True
        self.game_score = None
        common.GAME_OVER = False
        #
        # Cheating
        if self.options.cheat:
            serge.blocks.utils.addActorToWorld(
                world, serge.blocks.actors.FPSDisplay(G('fps-x'), G('fps-y'), G('fps-colour'), G('fps-size')),
                layer_name='debug')

    def createMinefield(self):
        """Create the minefield"""
        #
        difficulty = self.difficulties[self.difficulty_menu.getSelection()]
        #
        # The main minefield
        self.minefield = serge.blocks.utils.addActorToWorld(
            self.world,
            field.MineField('field', 'field',
                            difficulty[0], difficulty[1], difficulty[2],
                            difficulty[3] if common.C_GAME_MODE == common.C_MODE_NEW else 0),
            layer_name='main',
            center_position=G('field-position')
        )
        self.minefield.linkEvent(common.E_FLAG_CHANGED, self.flagChanged)
        self.minefield.linkEvent(common.E_GAME_OVER, self.gameOver)
        self.minefield.linkEvent(common.E_FIELD_CLEARED, self.fieldCleared)
        self.minefield.linkEvent(common.E_TRIGGER_HIT, self.triggerHit)

    def updateActor(self, interval, world):
        """Update this actor"""
        super(MainScreen, self).updateActor(interval, world)
        #
        # Update sensor
        square = self.minefield.getSquareAt(self.mouse.getScreenPos())
        if square:
            self.sensor.setValue(square.square.getSignalStrength() * 100.0 / 5.0)
            angles = [360-serge.blocks.directions.getAngleFromVector(v)
                      for v in square.square.getNeighbourSquareOffsets()]
            if angles:
                if self.options.cheat and self.keyboard.isShiftDown():
                    self.log.debug('Angle under cursor are %s' % (angles, ))
                self.scope.setAngles(angles)
            else:
                self.scope.setAngles([])
        else:
            self.sensor.setValue(0.0)
            self.scope.setAngles([])
        #
        if not common.GAME_OVER:
            #
            # Update timer
            self.clock_digits.setValue('%04d' % int(time.time() - self.start_time))
            #
            # Update the flag counter
            if common.GAME_TURNS > self._last_turns:
                #
                # We have cleared some cells since last time
                delta = common.GAME_TURNS - self._last_turns
                self.flag_counter.visual.setValue(
                    min(self.flag_counter.visual.getValue() + delta, G('flag-counter-max')))
                self._last_turns = common.GAME_TURNS
            elif time.time() - self._last_flag_placed >= G('flag-counter-decay-time'):
                #
                # No cells cleared and we can decay the counter
                self._last_flag_placed += G('flag-counter-decay-time')
                self.flag_counter.visual.setValue(max(0, self.flag_counter.visual.getValue() - 1))
        #
        if self.options.cheat:
            if self.keyboard.isClicked(pygame.K_n):
                self.fieldCleared(None, None)

    def showConfirmMenu(self, obj=None, arg=None):
        """Show the confirm menu"""
        serge.sound.Sounds.play('power')
        self.confirm.active = not self.confirm.active

    def flagChanged(self, change, arg):
        """Changed the number of flags"""
        self.total_flags -= change
        self.flags_digits.setValue('%04d' % self.total_flags)
        common.HAVE_FLAGS = self.total_flags > 0

    def gameOver(self, obj, arg):
        """The game is over"""
        common.GAME_OVER = True
        self.sensor.setValue(G('sensor-explosion'))
        self.game_over_elements.forEach().active = True
        self.success.visual.setText('Bad Luck!')
        self.navigation_lights.setStates(True)

    def restartClick(self, obj, arg):
        """The restart was clicked"""
        serge.sound.Sounds.play('select')
        if self.total_flags != self.getNumberBombs() and not common.GAME_OVER:
            self.confirm.active = True
        else:
            self.restartGame()

    def getNumberBombs(self, difficulty=None):
        """Return the number of bombs needed"""
        return self.difficulties[
            difficulty if difficulty else self.difficulty_menu.getSelection()][2]

    def restartGame(self):
        """Restart the game"""
        self.log.debug('Restarting game')
        self.world.removeActor(self.minefield)
        self.createMinefield()
        common.GAME_OVER = False
        common.HAVE_FLAGS = True
        common.GAME_TURNS = 0
        self.start_time = time.time()
        self.total_flags = self.getNumberBombs()
        self.flagChanged(0, None)
        self.game_over_elements.forEach().active = False
        self.game_score = None
        self.navigation_lights.setStates(False)
        self.flag_counter.active = self.scope.active = self._accepted_mode == common.C_MODE_NEW
        self._last_flag_placed = time.time()
        self._last_turns = 0

    def menuClick(self, menu, option):
        """The menu was clicked"""
        if self.initialised:
            serge.sound.Sounds.play('button')
            if self.total_flags != self.getNumberBombs(self._accepted_difficulty) and not common.GAME_OVER:
                self.confirm.active = True
            else:
                self.difficulty_lights.setStates(False)
                self._accepted_difficulty = option
                self.difficulty_lights.getButton(self._accepted_difficulty).setState(True)
                self.restartGame()

    def confirmClick(self, option):
        """The confirm menu was clicked"""
        serge.sound.Sounds.play('select')
        self.confirm.active = False
        if option == 'Start New Game':
            self.difficulty_lights.setStates(False)
            self._accepted_difficulty = self.difficulty_menu.getSelection()
            self._accepted_mode = self.mode_menu.getSelection()
            self.difficulty_lights.getButton(self._accepted_difficulty).setState(True)
            self.sensor.active = self._accepted_mode == common.C_MODE_NEW
            self.restartGame()
        elif option == 'Continue Current':
            self.difficulty_menu.selectItem(self._accepted_difficulty, do_callback=False)
            self.mode_menu.selectItem(self._accepted_mode, do_callback=False)
            self.mode_lights.setStates(False)
            self.mode_lights.getButton(self._accepted_mode).setState(True)
        else:
            self.engine.stop()

    def fieldCleared(self, obj, arg):
        """The field was cleared"""
        if not common.GAME_OVER:
            self.log.info('The field has been cleared')
            self.success.visual.setText('Well Done!')
            self.game_over_elements.forEach().active = True
            common.GAME_OVER = True
            self.game_score = time.time() - self.start_time
            self.navigation_lights.setStates(True)

    def gotoHighScores(self, obj, arg):
        """Go to the high scores screen"""
        serge.sound.Sounds.play('screen')
        world = self.engine.setCurrentWorldByName('high-score-screen')
        world.findActorByName('high-score-screen').setCurrentScore(
            self.game_score, self._accepted_difficulty, self._accepted_mode)

    def triggerHit(self, square, arg):
        """A trigger was hit"""
        self.log.info('Hit trigger on %s' % square.getNiceName())

    def setMode(self, mode, silent=False):
        """Set the game mode"""
        #
        if self.initialised:
            super(MainScreen, self).setMode(mode, silent)
            if self.total_flags != self.getNumberBombs(self._accepted_difficulty) and not common.GAME_OVER:
                self.confirm.active = True
            else:
                self._accepted_mode = mode
                self.sensor.active = mode == common.C_MODE_NEW
                self.restartGame()


def main(options):
    """Create the main logic"""
    #
    # The screen actor
    s = MainScreen(options)
    world = serge.engine.CurrentEngine().getWorld('main-screen')
    world.addActor(s)
    #
    # The behaviour manager
    manager = serge.blocks.behaviours.BehaviourManager('behaviours', 'behaviours')
    world.addActor(manager)
    #
    # Screenshots
    if options.screenshot:
        manager.assignBehaviour(None, 
            serge.blocks.behaviours.SnapshotOnKey(key=pygame.K_s, size=G('screenshot-size')
                , overwrite=False, location='screenshots'), 'screenshots')